/*
    Copyright (C) 2013 Rico Tzschichholz

    This file is part of Plank.

    Plank is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Plank is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http:  www.gnu.org/licenses/>.
*/

#ifdef TEST_DATA_DIR
#define DATA_DIR TEST_DATA_DIR
#else
#define DATA_DIR ""
#endif

#ifdef TEST_HOME_DIR
#define HOME_DIR TEST_HOME_DIR
#else
#define HOME_DIR ""
#endif
